-module(myapp_events3).
-behaviour(gen_server).

-export([start_link/0]).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-export([add_task/4, delete_task/2]).

-define(N_COUNT, 20_000).
-define(IS_STRING(S), (is_list(S) andalso S /= [] andalso is_integer(hd(S)))).


add_task(Task_Id, Proc_Id, Type, Time)
when ?IS_STRING(Task_Id), is_integer(Proc_Id), ((Type =:= static) orelse (Type =:= dynamic)), is_integer(Time) ->
  Time2 = case Type of
    dynamic ->
      erlang:system_time(second) + Time; % Timestamp_Now + Time
    _ ->
      % static
      Time
  end,
  % додаємо запис в СУБД
  % postgres("INSERT INTO events_table (task_id, proc_id, time) VALUES ($1, $2, $3)"
  %   " ON CONFLICT ON CONSTRAINT events_table_ids_uniq DO UPDATE SET time = $4, deleted = NULL",
  %   [Task_Id, Proc_Id, Time2, Time2]),
  
  ok.

delete_task(Task_Id, Proc_Id)
when ?IS_STRING(Task_Id), is_integer(Proc_Id) ->
  % позначаємо запис видаленим в СУБД
  % postgres("UPDATE events_table SET deleted = 1"
  %   " WHERE task_id = $1 AND proc_id = $2", [Task_Id, Proc_Id]),
  
  % видаляємо зі State (якщо є)
  gen_server:cast(?MODULE, {delete_task, {Task_Id, Proc_Id}}),
  ok.


start_link() ->
  gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).


init([]) ->
  ets:new(events_table, [ordered_set, named_table, { keypos, 1 }, private, {write_concurrency, true}]),
  
  self() ! check_time_process,
  State = [],
  {ok, State}.


handle_call(_Req, _From, State) ->
  {reply, not_handled, State}.


handle_cast({delete_task, {Task_Id, Proc_Id}=V}, State) ->
  % видаляємо зі State (якщо є)
  ets:delete(events_table, V),
  io:format("deleted: ~p ~p~n", [Task_Id, Proc_Id]),
  {noreply, State};


handle_cast(_Req, State) ->
  {noreply, State}.


handle_info(check_time_process, State) ->
  N0 = ets:info(events_table, 'size'),
  if N0 =:= 0 ; N0 < ?N_COUNT ->
      % отримаємо записи з СУБД
      N = ?N_COUNT - N0,
      ok = get_n_from_db(N),
      self() ! check_time_process;
      
    true ->
      % записи є, працюємо з ними
      erlang:send_after(500, self(), check_time_process),
      
      Timestamp_Now = erlang:system_time(second),
      %ets:fun2ms(fun({{A,B},C}) when C =< Timestamp_Now -> {{A,B},C} end).
      %[{{{'$1','$2'},'$3'}, [{'=<','$3',{const,1642507256}}], [{{{{'$1','$2'}},'$3'}}]}]
      
      L = ets:select(events_table, [{{{'$1','$2'},'$3'}, [{'=<','$3',{const, Timestamp_Now}}], [{{{{'$1','$2'}},'$3'}}]}] ),
      
      lists:foldl(fun({{Task_Id0, Proc_Id0}=V, _Time0}, ok) ->
        % send_to_rabbitmq({call_timer, Proc_Id0, Task_Id0}),
        
        % позначаємо запис видаленим в СУБД
        % postgres("UPDATE events_table SET deleted = 1"
        %   " WHERE task_id = $1 AND proc_id = $2", [Task_Id0, Proc_Id0]),
        
        ets:delete(events_table, V),
        io:format("tick: ~p ~p~n", [Task_Id0, Proc_Id0]),
        ok
      end, ok, L)
      
  end,
  {noreply, State};


handle_info(_Request, State) ->
  {noreply, State}.


terminate(_Reason, _State) ->
  ok.

code_change(_OldVsn, State, _Extra) -> 
  {ok, State}.


% отримати N записів з СУБД, додати в ETS
get_n_from_db(N) ->
  % отримуємо записи з СУБД
  % L = postgres("SELECT task_id, proc_id, time FROM events_table"
  %       " WHERE deleted IS NULL ORDER BY time DESC LIMIT $1", [N]),
  
  % додаємо отримані записи для обробки в ets
  %lists:foldl(fun({{_Task_Id0, _Proc_Id0}, _Time0}=V, ok) ->
  %   ets:insert(events_table, V),
  %   ok
  %  end, ok, L),
  
  ok.


